/**
 * CMI : Cluster Method Invocation
 * Copyright (C) 2007,2008 Bull S.A.S.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: ClientClusterViewManagerImpl.java 2038 2008-09-05 16:58:04Z loris $
 * --------------------------------------------------------------------------
 */

package org.ow2.cmi.controller.client;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Proxy;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.ejb.EJBHome;
import javax.ejb.EJBObject;

import net.jcip.annotations.ThreadSafe;

import org.ow2.cmi.admin.CMIMBeanConfigException;
import org.ow2.cmi.admin.MBeanUtils;
import org.ow2.cmi.config.CMIConfig;
import org.ow2.cmi.controller.common.AbsClusterViewManager;
import org.ow2.cmi.controller.common.CMIThreadFactory;
import org.ow2.cmi.controller.common.ClusterViewManager;
import org.ow2.cmi.controller.common.ImmutableClusteredObjectInfos;
import org.ow2.cmi.controller.provider.ClientClusterViewProvider;
import org.ow2.cmi.controller.provider.ClusteredClientClusterViewProvider;
import org.ow2.cmi.controller.server.ClusteredDummyRegistry;
import org.ow2.cmi.info.CMIInfoExtractor;
import org.ow2.cmi.info.CMIInfoExtractorException;
import org.ow2.cmi.info.ClusteredObjectInfo;
import org.ow2.cmi.lb.data.PolicyData;
import org.ow2.cmi.lb.policy.IPolicy;
import org.ow2.cmi.lb.strategy.IStrategy;
import org.ow2.cmi.pool.StubOrProxyFactory;
import org.ow2.cmi.reference.CMIReference;
import org.ow2.cmi.reference.CMIReferenceable;
import org.ow2.cmi.reference.ObjectNotFoundException;
import org.ow2.cmi.reference.ServerNotFoundException;
import org.ow2.cmi.reference.ServerRef;
import org.ow2.cmi.rpc.CMIProxy;
import org.ow2.cmi.rpc.POJOInvocationHandler;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.pool.api.IPoolConfiguration;
import org.ow2.util.pool.api.Pool;
import org.ow2.util.pool.impl.JPool;


/**
 * Manager that contains CMI logic for the client-side.
 * @author The new CMI team
 */
@ThreadSafe
public final class ClientClusterViewManagerImpl extends AbsClusterViewManager implements ClientClusterViewManager {

    /**
     * Logger.
     */
    private static final Log LOGGER = LogFactory.getLog(ClientClusterViewManagerImpl.class);

    /**
     * The InitialContextFactory name for constructing instances of the protocol-relative context.
     */
    private final String initialContextFactoryName;

    /**
     * The protocol name.
     */
    private final String protocol;

    /**
     * A daemon in charge of updating the cluster view.
     */
    private final Thread clientClusterViewUpdater;

    /**
     * Current selected ClientClusterViewProvider.
     */
    private static volatile ClientClusterViewProvider clientClusterViewProvider;

    /**
     * Cache of cluster view.
     */
    private final ClientClusterViewCache clientClusterViewCache = new ClientClusterViewCache();

    /**
     * Keep initial the server references of provider.
     */
    private List<CMIReference> providerServerRefs;

    /**
     * True if the cluster view provider is available.
     */
    private volatile boolean clusterViewProviderAvailable = false;

    private long configDate = -1;

    private volatile boolean refreshed;

    private IClientConfig configuration;

    /**
     * Builds a new manager for the client-side.
     * @param initialContextFactoryName an name of InitialContextFactory
     * @param protocol The protocol name
     */
    private ClientClusterViewManagerImpl(final String initialContextFactoryName, final String protocol, final List<ServerRef> serverRefs) {
        this.initialContextFactoryName = initialContextFactoryName;
        this.protocol = protocol;

        // Create the daemon in charge of updating the cluster view
        CMIThreadFactory cmiThreadFactory = getCmiThreadFactory();
        clientClusterViewUpdater = cmiThreadFactory.newThread(new Runnable() {

            public void run() {
                // Register this client (for statistics)
                try {
                    clientClusterViewProvider.registerClient(getUUID());
                } catch (Exception e) {
                    // No importance !
                    LOGGER.debug("Thread {0}> Cannot register the client", Thread.currentThread().getName(), e);
                }
                refreshed = true;
                do {
                    int refreshTime;
                    try {
                        refreshTime = getDelayToRefresh();
                    } catch (Exception e) {
                        // No importance !
                        refreshTime = 10000;
                    }
                    // Start latter because the system object is already up-to-date
                    LOGGER.debug("Thread {0} is sleeping for {1} millis...", Thread.currentThread().getName(), refreshTime);
                    try {
                        Thread.sleep(refreshTime);
                    } catch (InterruptedException e) {
                        // Restore interrupted status
                        Thread.currentThread().interrupt();
                    }
                    LOGGER.debug("Thread {0}> Updating the client cluster view...", Thread.currentThread().getName());
                    try {
                        updateClusterView();
                        if(getState().equals(State.STARTED)) {
                            setState(State.AVAILABLE);
                        }
                    } catch (Exception e) {
                        if(getState().equals(State.AVAILABLE)) {
                            setState(State.STARTED);
                        }
                        LOGGER.warn("Thread {0}> Cannot update the cluster view", Thread.currentThread().getName(), e);
                    }
                    // Now sleep...
                } while(refreshed);
            }
        }, "Cluster View Updater");

        // Set the default configuration
        configuration = new ClientConfig();

        // Initializes data of the provider of cluster view
        initDataOfProvider(serverRefs);

        // Initializes data of the dummy registry to enable a load-balancing of the JNDI accesses
        initDataOfDummyRegistry();
    }

    /**
     * Static method to get a singleton instance of ClientClusterViewManager.
     * @param initialContextFactoryName a name of class of InitialContextFactory
     * to use for constructing instances of the protocol-relative context
     * @param protocol a name of protocol
     * @param serverRefs a set of references on server
     * @return the instance (singleton) of ClientClusterViewManager
     * @throws ClientClusterViewManagerException if we cannot get an instance of ClientClusterViewManager.
     */
    public static synchronized ClientClusterViewManager getClientClusterViewManager(
            final String initialContextFactoryName, final String protocol, final List<ServerRef> serverRefs)
    throws ClientClusterViewManagerException {

        ClusterViewManager clusterViewManager = AbsClusterViewManager.getClusterViewManager();
        // If none instance of ClusterViewManager exists, we create one
        if(clusterViewManager == null) {
            LOGGER.debug("Constructing a ClientClusterviewManager...");
            clusterViewManager =
                new ClientClusterViewManagerImpl(initialContextFactoryName, protocol, serverRefs);

            AbsClusterViewManager.setClusterViewManager(clusterViewManager);
        } else {
            if(!(clusterViewManager instanceof ClientClusterViewManager)){
                 LOGGER.error("An instance of the manager that is not a ClientClusterViewManager already exists in the JVM");
                 throw new ClientClusterViewManagerException(
                         "An instance of the manager that is not a ClientClusterViewManager already exists in the JVM");
            }
        }
        return (ClientClusterViewManager) clusterViewManager;
    }

    /**
     * Creates a CMI proxy to enable invocations on the provider of cluster view.
     * @param serverRefs an initial set of the provider URLs for the provider of cluster view
     * @throws ClientClusterViewManagerException if a CMI proxy cannot be obtained for the provider of cluster view
     */
    private void initDataOfProvider(final List<ServerRef> serverRefs) throws ClientClusterViewManagerException {

        LOGGER.debug("**** Begin of initialization of the connection at provider...");
        String providerName = CMIConfig.getBindNameForProvider();

        // Extracts the informations about provider
        ClusteredObjectInfo clusteredObjectInfo;
        try {
            clusteredObjectInfo = CMIInfoExtractor.extractClusteringInfoFromAnnotatedPOJO(providerName,
                    ClientClusterViewProvider.class, ClusteredClientClusterViewProvider.class, false, false, null);
        } catch (CMIInfoExtractorException e) {
            LOGGER.error("Cannot extract the informations about provider", e);
            throw new ClientClusterViewManagerException("Cannot extract the informations about provider", e);
        }

        Class<?> itfClass = clusteredObjectInfo.getItfClass();
        Class<? extends EJBObject> businessClass = clusteredObjectInfo.getBusinessClass();
        String businessName = null;
        if(businessClass != null) {
            businessName = businessClass.getName();
        }
        ClientClusteredObjectView clusteredObjectView =
            clientClusterViewCache.addClusteredObject(
                    clusteredObjectInfo.getClusterName(), clusteredObjectInfo.isProxyEnabled(),
                    providerName, itfClass.getName(),
                    businessName, clusteredObjectInfo.hasState(),
                    clusteredObjectInfo.getApplicationExceptionNames());

        // Initializes the list of provider URL for getting the cluster view
        List<CMIReference> cmiReferences = new ArrayList<CMIReference>();
        for(ServerRef serverRef : serverRefs) {
            cmiReferences.add(new CMIReference(serverRef, providerName));
        }
        providerServerRefs = cmiReferences;
        clusteredObjectView.setCMIReferences(cmiReferences);

        // Initializes the cache
        clusteredObjectView.setItfClass(ClientClusterViewProvider.class);

        Class<? extends IPolicy<?>> policyClass = clusteredObjectInfo.getPolicyType();
        Class<? extends IStrategy<?>> strategyClass = clusteredObjectInfo.getStrategyType();

        String policyClassname = policyClass.getName();
        String strategyClassname = strategyClass.getName();

        PolicyData policyDataForProvider =
            new PolicyData(policyClassname, strategyClassname, clusteredObjectInfo.getProperties());
        clusteredObjectView.setPolicyData(policyDataForProvider);
        clientClusterViewCache.setPolicyClass(policyClassname, policyClass);
        clientClusterViewCache.setStrategyClass(strategyClassname, strategyClass);

        watch(providerName);
        try {
            updatePolicy(providerName);
        } catch (ObjectNotFoundException e) {
            LOGGER.error("Cannot initialize informations of the provider", e);
            throw new ClientClusterViewManagerException("Cannot initialize informations of the provider", e);
        }

        Pool<CMIReferenceable<?>, CMIReference> pool =
            new JPool<CMIReferenceable<?>, CMIReference>(new StubOrProxyFactory(this));
        pool.setPoolConfiguration(clusteredObjectInfo.getPoolConfiguration());
        setPool(providerName, pool);

        // Keep the current classloader: it can be a smart classloader
        ClassLoader classLoader = Thread.currentThread().getContextClassLoader();

        // Gets a CMI proxy to enable load-balanced calls at provider
        InvocationHandler invocationHandler =
             new POJOInvocationHandler(
                    classLoader, this, providerName, protocol, ClientClusterViewProvider.class);
        clientClusterViewProvider =
            (ClientClusterViewProvider) Proxy.newProxyInstance(
                    classLoader, new Class[] {ClientClusterViewProvider.class, CMIProxy.class}, invocationHandler);
        try {
            pullAndUpdateObjectInfos(providerName);
            clusterViewProviderAvailable = true;
        } catch (Exception e) {
            LOGGER.warn("Provider not found", e);
        }
        LOGGER.debug("**** End of initialization of the connection at provider.");
    }

    /**
     * Registers the dummy registry as an object to watch.
     */
    private void initDataOfDummyRegistry() {
        String dummyName = CMIConfig.getBindNameForDummyRegistry();
        ClusteredObjectInfo clusteredObjectInfo;
        // Extracts informations on clustering of dummy context
        try {
            clusteredObjectInfo =
                CMIInfoExtractor.extractClusteringInfoFromAnnotatedPOJO(
                        CMIConfig.getBindNameForDummyRegistry(), null,
                        ClusteredDummyRegistry.class, false, false, null);
        } catch (CMIInfoExtractorException e) {
            LOGGER.error("Cannot get infos for dummy registry", e);
            throw new ClientClusterViewManagerException("Cannot get infos for dummy registry", e);
        }
        clientClusterViewCache.addClusteredObject(
                clusteredObjectInfo.getClusterName(), clusteredObjectInfo.isProxyEnabled(),
                dummyName, null,
                null, clusteredObjectInfo.hasState(),
                clusteredObjectInfo.getApplicationExceptionNames());
        watch(dummyName);
        try {
            pullAndUpdateObjectInfos(dummyName);
        } catch (Exception e) {
            LOGGER.warn("Dummy registry not found", e);
        }
    }

    /**
     * Adds an object with the given name to a set of objects that need to have their state up-to-date.
     * @param objectName a name of object
     * @throws ObjectNotFoundException if none object has the given name
     * @throws ClientClusterViewManagerException if the name of interface of the object with the given name cannot be downloaded
     */
    @SuppressWarnings("unchecked")
    public void addObjectToWatch(final String objectName)
    throws ObjectNotFoundException, ClientClusterViewManagerException {
        if(!isWatched(objectName)) {
            LOGGER.debug("Initializing local data for object with name {0}", objectName);
            ImmutableClusteredObjectInfos immutableInfos;
            try {
                immutableInfos = clientClusterViewProvider.getImmutableInfos(objectName);
            } catch (RemoteException e) {
                LOGGER.error("Cannot get the immutable infos for {0}", objectName, e);
                throw new ClientClusterViewManagerException(
                        "Cannot get the immutable infos for " + objectName, e);
            }
            ClassLoader currentCLoader = Thread.currentThread().getContextClassLoader();

            synchronized (clientClusterViewCache) {
                ClientClusteredObjectView clusteredObjectView =
                    clientClusterViewCache.addClusteredObject(immutableInfos);

                String itfName = immutableInfos.getItfName();
                try {
                    Class<?> itfClass = Class.forName(itfName, true, currentCLoader);
                    clusteredObjectView.setItfClass(itfClass);
                } catch (ClassNotFoundException e) {
                    throw new ClientClusterViewManagerException(
                            "Cannot load the interface for name " + itfName, e);
                }
                // Check if the object is an ejb2
                if(Arrays.asList(getInterface(objectName).getInterfaces()).contains(EJBHome.class)) {
                    String businessName = immutableInfos.getBusinessName();

                    try {
                        Class<? extends EJBObject> businessClass =
                            (Class<? extends EJBObject>) Class.forName(businessName, true, currentCLoader);
                        clusteredObjectView.setBusinessClass(businessClass);
                    } catch (ClassNotFoundException e) {
                        throw new ClientClusterViewManagerException(
                                "Cannot load the business interface for name " + businessName, e);
                    }
                }
                watch(objectName);
                try {
                    pullAndUpdateObjectInfos(objectName);
                } catch (Exception e) {
                    LOGGER.warn("Cannot get object infos for {0}", objectName, e);
                }
            }
        }
    }

    public IClientConfig getConfig() {
        return configuration;
    }

    public boolean start() {
        if(isDestroyed()) {
            throw new IllegalStateException(
                    "This instance of CMI server can no more be used !!!");
        }
        if(getState().equals(State.STOPPED)) {
            setState(State.STARTING);
            // Start the daemon to update the cluster view
            clientClusterViewUpdater.start();

            if(CMIConfig.isEmbedded()) {
                // Registers CMIMBean
                try {
                    MBeanUtils.registerCMIMBean(this);
                } catch (CMIMBeanConfigException e) {
                    LOGGER.error("Cannot registers the CMI MBean", e);
                }
            }
            if(clusterViewProviderAvailable) {
                setState(State.AVAILABLE);
            } else {
                setState(State.STARTED);
            }
            LOGGER.info("The CMI client was successfully started.");
        }
        return clusterViewProviderAvailable;
    }

    public void stop() {
        if(getState().equals(State.STARTED) || getState().equals(State.AVAILABLE)) {
            setState(State.STOPPING);
            refreshed = false;
            if(CMIConfig.isEmbedded()) {
                // Unregister the CMI MBean
                try {
                    MBeanUtils.unregisterCMIMBean();
                } catch (CMIMBeanConfigException e) {
                    LOGGER.error("Cannot unregister CMIMBean", e);
                }
            }
            providerServerRefs = null;
            clientClusterViewCache.clear();
            clientClusterViewProvider = null;
            setState(State.STOPPED);
            LOGGER.info("The CMI client was successfully stopped.");
        }
    }


    /**
     * Returns the interface of an object bound with the given name.
     * @param objectName a name of object
     * @return the interface of an object bound with the given name
     * @throws ObjectNotFoundException if none object has the given name
     */
    public Class<?> getInterface(final String objectName) throws ObjectNotFoundException {
        try {
            return clientClusterViewCache.getClusteredObject(objectName).getItfClass();
        } catch (NullPointerException e) {
            LOGGER.error("Object is unknown: {0}", objectName);
            throw new ObjectNotFoundException("Object is unknown: " + objectName);
        }
    }

    /**
     * Returns the interface of an object bound with the given name.
     * @param objectName a name of object
     * @return the interface of an object bound with the given name
     * @throws ObjectNotFoundException if none object has the given name
     */
    public Class<? extends EJBObject> getRemoteClass(final String objectName) throws ObjectNotFoundException {
        try {
            return clientClusterViewCache.getClusteredObject(objectName).getBusinessClass();
        } catch (NullPointerException e) {
            LOGGER.error("Object is unknown: {0}", objectName);
            throw new ObjectNotFoundException("Object is unknown: " + objectName);
        }
    }

    /**
     * Returns the class that defines the policy for an object with the given name.
     * @param objectName a name of object
     * @return the class that defines the policy for an object with the given name
     * @throws ObjectNotFoundException if none object has the given name
     * @throws ClientClusterViewManagerException if the class of policy of the object with the given name cannot be downloaded
     */
    public Class<? extends IPolicy<?>> getPolicyClass(final String objectName)
    throws ObjectNotFoundException, ClientClusterViewManagerException {
        synchronized (clientClusterViewCache) {
            String classname = getPolicyClassName(objectName);
            Class<? extends IPolicy<?>> policy = clientClusterViewCache.getPolicyClass(classname);
            if(policy == null) {
                try {
                    policy = pullPolicyClass(objectName);
                } catch (Exception e) {
                    LOGGER.error("Cannot set the class of policy for {0}", objectName, e);
                    throw new ClientClusterViewManagerException("Cannot set the class of policy for " + objectName, e);
                }
                clientClusterViewCache.setPolicyClass(classname, policy);
            }
            return policy;
        }
    }

    /**
     * Returns the class that defines the strategy for an object with the given name.
     * @param objectName a name of object
     * @return the class that defines the strategy for an object with the given name
     * @throws ObjectNotFoundException if none object has the given name
     * @throws ClientClusterViewManagerException if the class of strategy of the object with the given name cannot be downloaded
     */
    public Class<? extends IStrategy<?>> getStrategyClass(final String objectName)
    throws ObjectNotFoundException, ClientClusterViewManagerException {
        synchronized (clientClusterViewCache) {
            String classname = getStrategyClassName(objectName);
            Class<? extends IStrategy<?>> strategy = clientClusterViewCache.getStrategyClass(classname);
            if(strategy == null) {
                try {
                    strategy = pullStrategyClass(objectName);
                } catch (Exception e) {
                    LOGGER.error("Cannot set the class of strategy for {0}", objectName, e);
                    throw new ClientClusterViewManagerException("Cannot set the class of strategy for " + objectName, e);
                }
                clientClusterViewCache.setStrategyClass(classname, strategy);
            }
            return strategy;
        }
    }

    /**
     * Returns a list of CMIReference for an object with the given name and protocol.
     * @param objectName a name of object
     * @param protocolName a name of protocol
     * @return a list of CMIReference for an object with the given name and protocol
     * @throws ObjectNotFoundException if none object has the given name for the given protocol
     */
    public List<CMIReference> getCMIReferences(final String objectName, final String protocolName)
    throws ObjectNotFoundException {
        try {
            return clientClusterViewCache.getClusteredObject(objectName).getCMIReferences();
        } catch (NullPointerException e) {
            LOGGER.error("Object is unknown: {0}", objectName);
            throw new ObjectNotFoundException("Object is unknown: " + objectName);
        }
    }

    /**
     * Returns the date of properties for an object with the given name.
     * @param objectName a name of object
     * @return the date of properties for an object with the given name
     * @throws ObjectNotFoundException if none object has the given name
     */
    public long getDateOfProperties(final String objectName) throws ObjectNotFoundException {
        try {
            return clientClusterViewCache.getClusteredObject(objectName).getPolicyData().getDateOfProperties();
        } catch (NullPointerException e) {
            LOGGER.error("Object is unknown: {0}", objectName);
            throw new ObjectNotFoundException("Object is unknown: " + objectName);
        }
    }

    /**
     * Returns the properties of the LB policy for an object with the given name.
     * @param objectName a name of object
     * @return the properties of the policy for an object with the given name
     * @throws ObjectNotFoundException if none object has the given name
     */
    public Map<String, Object> getPropertiesForPolicy(final String objectName) throws ObjectNotFoundException {
        try {
            return clientClusterViewCache.getClusteredObject(objectName).getPolicyData().getProperties();
        } catch (NullPointerException e) {
            LOGGER.error("Object is unknown: {0}", objectName);
            throw new ObjectNotFoundException("Object is unknown: " + objectName);
        }
    }

    /**
     * Returns a value property of the policy for an object with the given name.
     * @param objectName a name of object
     * @param propertyName a name of property
     * @return a value property of the policy for an object with the given name
     * @throws ObjectNotFoundException if none object has the given name
     */
    public Object getPropertyForPolicy(final String objectName, final String propertyName)
    throws ObjectNotFoundException {
        try {
            return clientClusterViewCache.getClusteredObject(objectName).getPolicyData().getProperties().get(propertyName);
        } catch (NullPointerException e) {
            LOGGER.error("Object is unknown: {0}", objectName);
            throw new ObjectNotFoundException("Object is unknown: " + objectName);
        }
    }

    /**
     * Returns a name of class that implements the interface InitialContextFactory for the given name of protocol.
     * @param protocolName a name of protocol
     * @return a name of class that implements the interface InitialContextFactory for the given name of protocol
     */
    public String getInitialContextFactoryName(final String protocolName) {
        return initialContextFactoryName;
    }

    /**
     * Returns the protocol used by this manager.
     * @return the protocol used by this manager
     */
    public Set<String> getProtocols() {
        return Collections.singleton(protocol);
    }

    /**
     * Returns the bytecode of the class with the given binary name.
     * @param binaryName a binary name of a class
     * @return the bytecode of the class with the given binary name
     * @throws ClientClusterViewManagerException if the bytecode cannot obtained
     */
    public byte[] getBytecode(final String binaryName) throws ClientClusterViewManagerException {
        try {
            return clientClusterViewProvider.getBytecode(binaryName);
        } catch (Exception e) {
            LOGGER.error("Cannot get bytecode for {0}", binaryName, e);
            throw new ClientClusterViewManagerException("Cannot get bytecode for " + binaryName, e);
        }
    }

    /**
     * Returns true if the object with the specified name is clustered.
     * @param objectName a name of object
     * @return true if the object with the specified name is clustered
     */
    public boolean isClustered(final String objectName) {
        try {
            return clientClusterViewProvider.isClustered(objectName);
        } catch (Exception e) {
            LOGGER.warn("Connection has failed: {0} non clustered", objectName, e);
            return false;
        }
    }

    /**
     * Returns the time between each update of the cluster view by clients.
     * @return the time between each update of the cluster view by clients
     * @throws ClientClusterViewManagerException if the delay between each update of the cluster view cannot be downloaded
     */
    public int getDelayToRefresh() throws ClientClusterViewManagerException {
        try {
            return clientClusterViewProvider.getDelayToRefresh();
        } catch (RemoteException e) {
            LOGGER.error("Cannot get delay to refresh", e);
            throw new ClientClusterViewManagerException("Cannot get delay to refresh", e);
        }
    }

    /**
     * Returns the name of class of policy for the object with the given name.
     * @param objectName a name of object
     * @return the name of class of policy for the object with the given name
     * @throws ObjectNotFoundException if none object has the given name
     */
    public String getPolicyClassName(final String objectName) throws ObjectNotFoundException {
        try {
            return clientClusterViewCache.getClusteredObject(objectName).getPolicyData().getPolicyType();
        } catch (NullPointerException e) {
            LOGGER.error("Object is unknown: {0}", objectName);
            throw new ObjectNotFoundException("Object is unknown: " + objectName);
        }
    }

    /**
     * Returns the name of class of strategy for the object with the given name.
     * @param objectName a name of object
     * @return the name of class of strategy for the object with the given name
     * @throws ObjectNotFoundException if none object has the given name
     */
    public String getStrategyClassName(final String objectName) throws ObjectNotFoundException {
        try {
            return clientClusterViewCache.getClusteredObject(objectName).getPolicyData().getStrategyType();
        } catch (NullPointerException e) {
            LOGGER.error("Object is unknown: {0}", objectName);
            throw new ObjectNotFoundException("Object is unknown: " + objectName);
        }
    }

    /**
     * Returns the name of cluster for the object with the given name.
     * @param objectName a name of object
     * @return the name of cluster for a object with the given name
     * @throws ObjectNotFoundException if none object has the given name
     * @throws ClientClusterViewManagerException if the name of the cluster, that contains the object with the given name,
     * cannot be downloaded
     */
    public String getClusterName(final String objectName) throws ObjectNotFoundException {
        try {
            return clientClusterViewCache.getClusteredObject(objectName).getClusterName();
        } catch (NullPointerException e) {
            LOGGER.error("Object is unknown: {0}", objectName);
            throw new ObjectNotFoundException("Object is unknown: " + objectName);
        }
    }

    public boolean isProxyEnabled(final String objectName) throws ObjectNotFoundException {
        try {
            return clientClusterViewCache.getClusteredObject(objectName).isProxyEnabled();
        } catch (NullPointerException e) {
            LOGGER.error("Object is unknown: {0}", objectName);
            throw new ObjectNotFoundException("Object is unknown: " + objectName);
        }
    }

    /**
     * Returns the load-factor for the server with the given address.
     * @param serverRef a reference on a server
     * @return the load-factor for the server with the given address
     */
    public int getLoadFactor(final ServerRef serverRef) {
        return clientClusterViewCache.getLoadFactor(serverRef);
    }

    /**
     * Return true if the object with the given name is stateful.
     * @param objectName a name of object
     * @return true if the object with the given name is stateful
     * @throws ObjectNotFoundException if none object has the given name
     */
    public boolean hasState(final String objectName) throws ObjectNotFoundException {
        try {
            return clientClusterViewCache.getClusteredObject(objectName).isStateful();
        } catch (NullPointerException e) {
            LOGGER.error("Object is unknown: {0}", objectName);
            throw new ObjectNotFoundException("Object is unknown: " + objectName);
        }
    }

    /**
     * Return true if the object with the given name is replicated for high-availability.
     * @param objectName a name of object
     * @return true if the object with the given name is replicated for high-availability
     * @throws ObjectNotFoundException if none object has the given name
     */
    public boolean isReplicated(final String objectName)
    throws ObjectNotFoundException {
        try {
            return clientClusterViewCache.getClusteredObject(objectName).isReplicated();
        } catch (NullPointerException e) {
            LOGGER.error("Object is unknown: {0}", objectName);
            throw new ObjectNotFoundException("Object is unknown: " + objectName);
        }
    }

    /**
     * Return classnames of the application exceptions.
     * @param objectName a name of object
     * @return classnames of the application exceptions
     * @throws ObjectNotFoundException if none object has the given name
     */
    public Set<String> getApplicationExceptionNames(final String objectName) throws ObjectNotFoundException{
        try {
            return clientClusterViewCache.getClusteredObject(objectName).getApplicationExceptionNames();
        } catch (NullPointerException e) {
            LOGGER.error("Object is unknown: {0}", objectName);
            throw new ObjectNotFoundException("Object is unknown: " + objectName);
        }
    }

    /**
     * @return the set of clustered object names
     */
    public Set<String> getObjectNames() {
        return clientClusterViewCache.getObjectNames();
    }

    /**
     * Returns true if the pool for object with the given name should be empty.
     * @param objectName a name of object
     * @return true if the pool for object with the given name should be empty
     */
    public boolean isPoolToEmpty(final String objectName) {
        // TODO
        return false;
    }

    /**
     * Returns the configuration of pool of CMIReferenceable for a object with the given name.
     * @param objectName a name of object
     * @return the configuration of pool of CMIReferenceable for a object with the given name
     * @throws ObjectNotFoundException
     */
    public IPoolConfiguration getPoolConfiguration(final String objectName) throws ObjectNotFoundException {
        try {
            return clientClusterViewProvider.getPoolConfiguration(objectName);
        } catch (RemoteException e) {
            LOGGER.error("Cannot get the configuration of pool for object with name {0}", objectName, e);
            throw new ClientClusterViewManagerException(
                    "Cannot get the configuration of pool for object with name " + objectName, e);
        }
    }

    /************* Start of methods to pull the cluster view: *****************/

    /**
     * Updates informations about clustering for every watched objects.
     * @throws ClientClusterViewManagerException if the update failed
     * @throws RemoteException
     */
    public void updateClusterView()
    throws ClientClusterViewManagerException, RemoteException {

        pullAndUpdateConfiguration();

        // Update list of node for each clustered RMI object
        for(String objectName : getNamesOfWatchedObject()) {
            try {
                pullAndUpdateObjectInfos(objectName);
            } catch (ObjectNotFoundException e) {
                LOGGER.error("The object with name {0} is unknown", objectName, e);
                throw new ClientClusterViewManagerException(
                        "The object with name " + objectName + " is unknown", e);
            }
        }
        pullAndUpdateLoadFactors();
    }

    private void pullAndUpdateConfiguration() throws RemoteException {
        long configDate = clientClusterViewProvider.getDateOfConfiguration();
        if(configDate > this.configDate) {
            setConfiguration(clientClusterViewProvider.getConfiguration());
            this.configDate = configDate;
        }
    }

    private void setConfiguration(final IUpdatedClientConfig updatedClientConfig) {
        configuration.setUpdatedConfig(updatedClientConfig);
    }

    /**
     * Pulls the Load Balance policy class.
     * @param objectName a name of object
     * @return A Load Balance policy class
     * @throws ObjectNotFoundException if none object has the given name
     * @throws RemoteException if the class of policy of the object with the given name cannot be downloaded
     */
    private Class<? extends IPolicy<?>> pullPolicyClass(final String objectName)
    throws ObjectNotFoundException, RemoteException {
        // Set classloader to download class of policy
        ClassLoader oldClassLoader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(new CMIClientClassLoader(oldClassLoader, this));
        Class<? extends IPolicy<?>> policyClass;
        try {
            policyClass = clientClusterViewProvider.getPolicyClass(objectName);
        } catch(ClassNotFoundException e) {
            LOGGER.error("Cannot download the policy class for {0}", objectName, e);
            throw new ClientClusterViewManagerException(
                    "Cannot download the policy class for " + objectName, e);
        } finally {
            Thread.currentThread().setContextClassLoader(oldClassLoader);
        }
        return policyClass;
    }

    /**
     * Pulls the Load Balance strategy class.
     * @param objectName a name of object
     * @return A Load Balance strategy class
     * @throws ObjectNotFoundException if none object has the given name
     * @throws RemoteException if the class of strategy of the object with the given name cannot be downloaded
     */
    private Class<? extends IStrategy<?>> pullStrategyClass(final String objectName)
    throws RemoteException, ObjectNotFoundException {
        // Set classloader to download class of strategy
        ClassLoader oldClassLoader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(new CMIClientClassLoader(oldClassLoader, this));
        Class<? extends IStrategy<?>> strategyClass;
        try {
            strategyClass = clientClusterViewProvider.getStrategyClass(objectName);
        } catch(ClassNotFoundException e) {
            LOGGER.error("Cannot download the strategy class for {0}", objectName, e);
            throw new ClientClusterViewManagerException(
                    "Cannot download the strategy class for " + objectName, e);
        } finally {
            Thread.currentThread().setContextClassLoader(oldClassLoader);
        }
        return strategyClass;
    }

    /**
     * Updates informations about clustering for the object with the given name.
     * @param objectName a name of object
     * @throws ObjectNotFoundException if none object has the given name
     * @throws ClientClusterViewManagerException if the update failed
     */
    public void pullAndUpdateObjectInfos(final String objectName)
    throws RemoteException, ObjectNotFoundException {
        LOGGER.debug("Updating infos for: {0} - thread is {1}", objectName, Thread.currentThread().getName());

        synchronized (clientClusterViewCache) {
            if(objectName.equals(CMIConfig.getBindNameForProvider())
                    && clientClusterViewCache.getClusteredObject(objectName).getCMIReferences().isEmpty()) {
                LOGGER.warn("No provider URLs found for the provider, trying with the initials: ", providerServerRefs);
                clientClusterViewCache.getClusteredObject(objectName).setCMIReferences(providerServerRefs);
            }
            pullAddUpdateReplicationInfos(objectName);
            pullAndUpdatePolicy(objectName);
            pullAndUpdateCMIReferences(objectName);
        }
    }

    private void pullAddUpdateReplicationInfos(final String objectName)
    throws RemoteException, ObjectNotFoundException {
        clientClusterViewCache.getClusteredObject(objectName).setReplicated(
                clientClusterViewProvider.isReplicated(objectName));
    }

    /**
     * Pulls and updates the policy for the object with the given name.
     * @param objectName a name of object
     * @throws RemoteException if the information about policy cannot be downloaded
     * @throws ObjectNotFoundException
     */
    private void pullAndUpdatePolicy(final String objectName)
    throws RemoteException, ObjectNotFoundException {
        // Pulls the name of class for policy
        String lastPolicyName = clientClusterViewProvider.getPolicyClassName(objectName);
        LOGGER.debug("New policy name: {0}", lastPolicyName);

        // Pulls the name of class for strategy
        String lastStrategyName = clientClusterViewProvider.getStrategyClassName(objectName);
        LOGGER.debug("New strategy name: {0}", lastStrategyName);

        // Pulls the date of properties
        long dateOfLastProps = clientClusterViewProvider.getDateOfProperties(objectName);
        LOGGER.debug("New date of properties: {0}", dateOfLastProps);

        // Compares with the current data
        if(clientClusterViewCache.getClusteredObject(objectName).getPolicyData() != null) {
            // Cache contains some data
            String policyName = getPolicyClassName(objectName);
            LOGGER.debug("Old policy name: {0}", policyName);
            long dateOfProps = getDateOfProperties(objectName);
            LOGGER.debug("Old date of properties: {0}", dateOfProps);
            String strategyName = getStrategyClassName(objectName);
            LOGGER.debug("Old strategy name: {0}", strategyName);
            if(policyName.equals(lastPolicyName)
                    && dateOfProps == dateOfLastProps
                    && strategyName.equals(lastStrategyName)) {
                return;
            }
        }
        Map<String, Object> lastProperties;

        if(dateOfLastProps == 0) {
            // No property
            lastProperties = new HashMap<String, Object>();
        } else {
            lastProperties = clientClusterViewProvider.getPropertiesForPolicy(objectName);
        }
        clientClusterViewCache.getClusteredObject(objectName).setPolicyData(new PolicyData(
                lastPolicyName, lastStrategyName, lastProperties, dateOfLastProps));
        updatePolicy(objectName);
    }

    /**
     * Pulls and updates the CMI references for the object with the given name.
     * @param objectName a name of object
     * @throws RemoteException
     * @throws ObjectNotFoundException
     */
    private void pullAndUpdateCMIReferences(final String objectName)
    throws RemoteException, ObjectNotFoundException {
        // Update the list of nodes for this RMI object
        Collection<CMIReference> cmiReferences = clientClusterViewProvider.getCMIReferences(objectName, protocol);
        // Update the load-factor of each server
        for(CMIReference cmiRef : cmiReferences) {
            pullAndUpdateMissingLoadFactor(cmiRef.getServerRef());
        }
        clientClusterViewCache.getClusteredObject(objectName).setCMIReferences(cmiReferences);
    }

    private void pullAndUpdateMissingLoadFactor(final ServerRef serverRef) throws RemoteException {
        try {
            clientClusterViewCache.getLoadFactor(serverRef);
        } catch(NullPointerException npe) {
            int loadFactor;
            try {
                loadFactor = clientClusterViewProvider.getLoadFactor(serverRef);
                clientClusterViewCache.setLoadFactor(serverRef, loadFactor);
            } catch (ServerNotFoundException e) {
                LOGGER.warn("Server not found: {0}", serverRef, e);
            }
        }
    }

    /**
     * Pulls and updates the load-factors of the server known by this client.
     * @throws RemoteException
     */
    private void pullAndUpdateLoadFactors() throws RemoteException {
        for(ServerRef address : clientClusterViewCache.getAddressesOfServer()) {
            try {
                int loadFactor = clientClusterViewProvider.getLoadFactor(address);
                clientClusterViewCache.setLoadFactor(address, loadFactor);
            } catch (ServerNotFoundException e) {
                LOGGER.debug("server not found: {0}", address, e);
                clientClusterViewCache.removeLoadFactor(address);
            }
        }
    }

    /************* End of methods to pull the cluster view *****************/

}
